' //////////////////////////////////////////////////////////////////////
' eeprom
' 
' AUTHOR: Joshua Meeds
' LAST MODIFIED: 3-12-2007
'
' //////////////////////////////////////////////////////////////////////

'///////////////////////////////////////////////////////////////////////
' CONSTANTS SECTION ////////////////////////////////////////////////////
'///////////////////////////////////////////////////////////////////////

CON

  SCL         = 28
  SDA         = 29

  WRITEBUFFER_SIZE  = 128
  LINE_SIZE   = 32

  FUNCTION_START_PAGE = 129
 
  READ        = %10100001
  WRITE       = %10100000

'///////////////////////////////////////////////////////////////////////
' VARIABLES SECTION ////////////////////////////////////////////////////
'///////////////////////////////////////////////////////////////////////

VAR
  byte turbo_page
  
'///////////////////////////////////////////////////////////////////////
' EXPORT PUBLICS  //////////////////////////////////////////////////////
'///////////////////////////////////////////////////////////////////////

''---------------------------------------------------------------------
''ResetEeprom
''Purpose: Make sure EEPROM is in a good state
''Params: None
''Returns:nothing
''---------------------------------------------------------------------
PUB ResetEeprom | counter
   dira[SCL] := 1
   dira[SDA] := 0
   outa[SCL] := 1
   counter := 0
   repeat
      outa[SCL] := 0
      outa[SCL] := 1
      counter++
   while ((counter < 9) and (ina[SDA] == 0))

  outa[SCL] := 0
  outa[SDA] := 0
  dira[SDA] := 1
  outa[SCL] := 1     ' Drive the clock, data high
  outa[SDA] := 1
   

''---------------------------------------------------------------------
''WritePages
''Purpose: Write a number of complete 256-byte pages
''Params: SourceAddress,   memory we are writing from
''        DestinationPage, which page (of 256) we are writing to
''        NumberOfPages,   how many full pages to write
''Returns:nothing
''---------------------------------------------------------------------
PUB WritePages(SourceAddress, DestinationPage, NumberOfPages) | data, page, pagebackup
  pagebackup := DestinationPage
  Start(WRITE)
  
  repeat NumberOfPages
    page := DestinationPage

    ' Send the page
    repeat 8
      outa[SDA] := (page & %10000000) >> 7     ' Set a bit
      outa[SCL] := 1                                      ' Clock it in                                      
      outa[SCL] := 0
      page <<= 1                               ' Prepare the next bit

    Check

    ' Address zero (beginning of page)
    repeat 8
      outa[SDA] := 0
      outa[SCL] := 1                               ' Clock it in                                      
      outa[SCL] := 0

    Check

    ' And finally, write the data
    repeat WRITEBUFFER_SIZE
      data := byte[SourceAddress]
      repeat 8
        outa[SDA] := (data & %10000000) >> 7     ' Set a bit
        outa[SCL] := 1                           ' Clock it in                                      
        outa[SCL] := 0
        data <<= 1                               ' Prepare the next bit

      Check
      SourceAddress++

    Stop
          
    repeat while (Start(WRITE) <> 0)

'************Do it again for the second half**********

    page := DestinationPage

    ' Send the page
    repeat 8
      outa[SDA] := (page & %10000000) >> 7     ' Set a bit
      outa[SCL] := 1                                      ' Clock it in                                      
      outa[SCL] := 0
      page <<= 1                               ' Prepare the next bit

    Check

    ' Address 128 (halfway through), send a 1 and seven 0's
    outa[SDA] := 1
    outa[SCL] := 1                               ' Clock it in                                      
    outa[SCL] := 0
    repeat 7
      outa[SDA] := 0
      outa[SCL] := 1                               ' Clock it in                                      
      outa[SCL] := 0

    Check

    ' And finally, write the data
    repeat WRITEBUFFER_SIZE
      data := byte[SourceAddress]
      repeat 8
        outa[SDA] := (data & %10000000) >> 7     ' Set a bit
        outa[SCL] := 1                           ' Clock it in                                      
        outa[SCL] := 0
        data <<= 1                               ' Prepare the next bit

      Check
      SourceAddress++

    Stop
          
    repeat while (Start(WRITE) <> 0)

    DestinationPage++

  if ((turbo_page <> pagebackup) and (pagebackup => FUNCTION_START_PAGE))
    SourceAddress -= 512
    ReadLines(0, turbo_page,   0, SourceAddress, 8)
    ReadLines(0, turbo_page+1, 0, SourceAddress+256, 8)

  return
' WritePages

''---------------------------------------------------------------------
''ReadLines
''Purpose: Read 32 bytes (one line) from an EEPROM page
''Params: TurboBuffer, RAM Buffer to read turbo functions from
''        SourcePage,  which page (of 256) we are reading from
''        SourceLine,  which line (of 8) we are reading from
''        DestinationAddress,  memory we are reading into
''Returns:nothing
''---------------------------------------------------------------------
PUB ReadLines(TurboBuffer, SourcePage, SourceLine, DestinationAddress, NumberOfLines)
  SourceLine *= LINE_SIZE
  NumberOfLines *= LINE_SIZE

  if ( (TurboBuffer > 0) and ((turbo_page == SourcePage) or (turbo_page+1 == SourcePage)) )
  ' Turbo read!  Read from speedy RAM rather than slow EEPROM
    if (turbo_page+1 == SourcePage)
      TurboBuffer += 256    
    repeat NumberOfLines
      byte[DestinationAddress] := byte[TurboBuffer+SourceLine]
      DestinationAddress++
      TurboBuffer++
    return
  
'  Start(WRITE)      ' no need to wait for writes, this will never be run just after a write
  repeat while (Start(WRITE) <> 0)

  ' Send the page
  repeat 8
    outa[SDA] := (SourcePage & %10000000) >> 7     ' Set a bit
    outa[SCL] := 1                                 ' Clock it in                                      
    outa[SCL] := 0
    SourcePage <<= 1                               ' Prepare the next bit

  Check

  ' Send the address of the line
  repeat 8
    outa[SDA] := (SourceLine & %10000000) >> 7     ' Set a bit
    outa[SCL] := 1                                 ' Clock it in                                      
    outa[SCL] := 0
    SourceLine <<= 1                               ' Prepare the next bit

  Check

  Start(READ)

  ' And finally, read the data
  dira[SDA] := 0       ' Set pin to Input
  
  repeat NumberOfLines-1
    byte[DestinationAddress] := 0
    repeat 8
      byte[DestinationAddress] <<= 1
      outa[SCL] := 1                                                               
      byte[DestinationAddress] |= ina[SDA]
      outa[SCL] := 0

    outa[SDA] := 0            ' Send our own Check bit back (we got it!)
    dira[SDA] := 1
    outa[SCL] := 1            ' Clock it out
    outa[SCL] := 0
    outa[SDA] := 0            ' Prepare for the next read
    dira[SDA] := 0
    
    DestinationAddress++

  byte[DestinationAddress] := 0
  repeat 8
    byte[DestinationAddress] <<= 1
    outa[SCL] := 1                                                               
    byte[DestinationAddress] |= ina[SDA]
    outa[SCL] := 0

  outa[SDA] := 1            ' Cancel the read
  dira[SDA] := 1
  outa[SCL] := 1            ' Clock it out
  outa[SCL] := 0
  outa[SDA] := 0            ' Prepare for the next read

  Stop

  return
' ReadLines

''---------------------------------------------------------------------
''SetTurboPage
''Purpose: Set an EEPROM page as a "turbo page" to read from RAM rather than EEPROM
''Params: Page - Page (of 256) to set as Turbo
''Returns:nothing
''---------------------------------------------------------------------
PUB SetTurboPage(Page, SourceAddress)
  if ((turbo_page <> Page) and (SourceAddress > 0))
    ReadLines(0, Page,   0, SourceAddress, 8)
    ReadLines(0, Page+1, 0, SourceAddress+256, 8)
  turbo_page := Page
  return
' SetTurboPage

'///////////////////////////////////////////////////////////////////////
' LOCAL PRIVATES  //////////////////////////////////////////////////////
'///////////////////////////////////////////////////////////////////////

''---------------------------------------------------------------------
''Start
''Purpose: Tell the Hydra EEPROM to get ready for reads/writes
''Params: Direction - 8-bit Start address with direction-bit set
''Returns:nothing
''---------------------------------------------------------------------
PRI Start(Direction) | bitTest
  dira[SCL] := 1     ' Set to output
  dira[SDA] := 1
  outa[SCL] := 1     ' Drive the clock, data high
  outa[SDA] := 1
  
  outa[SDA] := 0     ' Set data to low while clock is still high
  outa[SCL] := 0     ' Set clock low

  repeat 8                                        ' Send the 8-bit Device Address for this direction
    outa[SDA] := (Direction & %10000000) >> 7     ' Set a bit
    outa[SCL] := 1                                ' Clock it in                                      
    outa[SCL] := 0
    Direction <<= 1                               ' Prepare the next bit

  bitTest := Check

  return (bitTest)
' Start

''---------------------------------------------------------------------
''Stop
''Purpose: Tell the Hydra EEPROM to stop reading/writing
''Params: none
''Returns:nothing
''---------------------------------------------------------------------
PRI Stop
  outa[SCL] := 1     ' Drive the clock, data high
  outa[SDA] := 1     ' Set data from low to high while the clock is still high
  outa[SCL] := 0     ' Set clock low

  return
' Stop

''---------------------------------------------------------------------
''Check
''Purpose: Every read/write, the EEPROM returns a zero.  Check it.
''Params: none
''Returns:The bit returned by the EEPROM
''---------------------------------------------------------------------
PRI Check : CheckReturn
  dira[SDA] := 0           ' Set to Input
  outa[SCL] := 1     
  CheckReturn := ina[SDA]  ' Clock the Check bit out
  outa[SCL] := 0
  outa[SDA] := 0           ' Prepare for the next read/write
  dira[SDA] := 1

  return(CheckReturn)
' Check

'///////////////////////////////////////////////////////////////////////
' DATA SECTION /////////////////////////////////////////////////////////
'///////////////////////////////////////////////////////////////////////